/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.onlineSim.cac;

import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.interfaces.simulation.IEventProcessor;
import com.net2plan.interfaces.simulation.SimAction;
import com.net2plan.interfaces.simulation.SimEvent;
import com.net2plan.libraries.CandidatePathList;
import com.net2plan.libraries.WDMUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.StringUtils;
import com.net2plan.utils.Triple;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * <p>This algorithm receives lightpath connection requests and releases. For each 
 * lightpath release, wavelength resources are liberated. For each lightpath request, 
 * it determines how many lightpaths are actually required. Then, it tries to allocate 
 * them over a set of precomputed all-pairs <tt>k</tt>-shortest paths (maximum path length in km 
 * equal to <tt>maxLightpathLengthInKm</tt>) using first-fit wavelength assignment 
 * (wavelength continuity is enforced).</p>
 * 
 * <p>Let \( trafficVolume \) be the traffic volume request for the connection 
 * request, and \( binaryRatePerChannel_Gbps \) the binary rate per lightpath. 
 * Then, the number of lightpaths, denoted as \( numLps \), to be established 
 * is computed as follows:</p>
 * 
 * <p><i>numLps = ceil(trafficVolume/binaryRatePerChannel_Gbps</i></p>
 * 
 * <p>Then, all the lightpaths are tried to be allocated as aforementioned. If 
 * there are no resources for all lightpaths, connection may be accepted partially.</p>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class CACSim_AA_WDM_basicAlgorithm extends IEventProcessor
{
	private Map<Long, Pair<Long, Long>> fiberMap;
	private CandidatePathList cpl;
	
	private double binaryRatePerChannel_Gbps;
	private Map<Long, Integer> w_f;
	private List<Set<Long>> wavelengthFiberOccupancy;

	@Override
	public String getDescription()
	{
		String newLine = StringUtils.getLineSeparator();

		StringBuilder description = new StringBuilder();
		description.append("This algorithm receives lightpath connection requests and releases. For each "
				+ "lightpath release, wavelength resources are liberated. For each lightpath request, "
				+ "it determines how many lightpaths are actually required. Then, it tries to allocate "
				+ "them over a set of precomputed all-pairs k-shortest paths (maximum path length in km "
				+ "equal to maxLightpathLengthInKm) using first-fit wavelength assignment (wavelength continuity is enforced)");
		description.append(newLine).append(newLine);
		description.append("If there are no resources for all lightpaths, connection may be accepted partially");

		return description.toString();
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("binaryRatePerChannel_Gbps", "10", "Binary rate of all the lightpaths"));
		algorithmParameters.add(Triple.of("maxLightpathLengthInKm", "5000", "Maximum allowed lightpath length in km"));
		algorithmParameters.add(Triple.of("k", "3", "Maximum number of candidate paths per demand"));

		return algorithmParameters;
	}

	@Override
	public void initialize(NetPlan initialNetPlan, NetPlan currentNetPlan, Map<String, String> algorithmParameters, Map<String, String> simulationParameters, Map<String, String> net2planParameters)
	{
		WDMUtils.checkConsistency(currentNetPlan);
		
		int k = Integer.parseInt(algorithmParameters.get("k"));
		double maxLightpathLengthInKm = Double.parseDouble(algorithmParameters.get("maxLightpathLengthInKm"));

		binaryRatePerChannel_Gbps = Double.parseDouble(algorithmParameters.get("binaryRatePerChannel_Gbps"));
		cpl = new CandidatePathList(initialNetPlan, "maxLengthInKm", Double.toString(maxLightpathLengthInKm), "K", Integer.toString(k));
		fiberMap = currentNetPlan.getLinkMap();
		w_f = WDMUtils.getFiberNumWavelengthsMap(currentNetPlan);
		wavelengthFiberOccupancy = WDMUtils.getWavelengthFiberOccupancy(currentNetPlan);
	}

	@Override
	public void processEvent(NetPlan currentNetPlan, SimEvent event)
	{
		List<SimAction> actions = event.getEventActionList();
		
		for(SimAction action : actions)
		{
			switch(action.getActionType())
			{
				case DEMAND_ADDED:
				{
					long lpDemandId = action.getDemandAdded_demandId();
					long originalDemandId = Long.parseLong(currentNetPlan.getDemandAttribute(lpDemandId, "originalDemandId"));

					double h_d = currentNetPlan.getDemandOfferedTraffic(lpDemandId);
					int numLps = (int) Math.ceil(h_d / binaryRatePerChannel_Gbps);

					for (int lp = 0; lp < numLps; lp++)
					{
						boolean failToAllocate = true;

						for (long pathId : cpl.getDemandPaths(originalDemandId))
						{
							List<Long> seqFibers_thisLp = cpl.getPathSequenceOfLinks(pathId);
							int[] seqWavelengths_thisLp = WDMUtils.WA_firstFit(fiberMap, seqFibers_thisLp, w_f, wavelengthFiberOccupancy);
							if (seqWavelengths_thisLp.length == 0) continue;

							failToAllocate = false;

							/* add this route, updating wavelength occupation */
							long lpId = currentNetPlan.addRoute(lpDemandId, binaryRatePerChannel_Gbps, seqFibers_thisLp, null);
							WDMUtils.allocateResources(seqFibers_thisLp, seqWavelengths_thisLp, wavelengthFiberOccupancy);
							WDMUtils.setLightpathSeqWavelengths(currentNetPlan, lpId, seqWavelengths_thisLp);
							break;
						}

						if (failToAllocate == true) break;
					}

					break;
				}
				
				case ROUTE_REMOVED:
				{
					/* update wavelength occupation */
					List<Long> seqFibers_thisLp = action.getRouteRemoved_sequenceOfLinks();
					int[] seqWavelengths_thisLp = StringUtils.toIntArray(StringUtils.split(action.getRouteRemoved_attributes().get(WDMUtils.SEQUENCE_OF_WAVELENGTHS_ATTRIBUTE_NAME), " "));
					WDMUtils.releaseResources(seqFibers_thisLp, seqWavelengths_thisLp, wavelengthFiberOccupancy);
					
					break;
				}
			}
		}
	}
}
